{$I jedi.inc}
{$IFDEF COMPILER6_UP}
  {$WARN SYMBOL_DEPRECATED OFF}
{$ENDIF}
{$IFDEF COMPILER7_UP}
  {$WARN SYMBOL_DEPRECATED OFF}
  {$WARN UNSAFE_CODE OFF}
  {$WARN UNSAFE_TYPE OFF}
  {$WARN UNSAFE_CAST OFF}
{$ENDIF}
{$ALIGN ON}
{$MINENUMSIZE 4}

unit StreamFilter;

interface

uses
  SysUtils, Classes, DSPack, ActiveX, DirectShow9, DirectDraw, DSUtil, SyncObjs, Direct3D9, WMF9,
  BaseClass, UAsyncIO, Types, UAsyncRdr, dialogs;

type
  TMyFilterGraph=class(TFilterGraph)
  public
    {@exclude}
    procedure InsertFilter(AFilter: IFilter);
    {@exclude}
    procedure RemoveFilter(AFilter: IFilter);
  end;

  TBCTStream=class(TBCAsyncStream)
  private
    FLock: TBCCritSec;
    FStream:TStream;
    FOffset:integer;
  public
    constructor Create(AStream: TStream; AOffset:integer);
    destructor Destroy; override;
    function SetPointer(APos: LONGLONG): HResult; override;
    function Read(ABuffer: PByte; ABytesToRead: DWord;
      AAlign: Boolean; out ABytesRead: DWord): HResult; override;
    function Size(out ASizeAvailable: LONGLONG): LONGLONG; override;
    function Alignment: DWord; override;
    procedure Lock; override;
    procedure Unlock; override;
  end;

  TBCTStreamReader=class(TBCAsyncReader)
  public
    //  We're not going to be CoCreate'd so we don't need registration
    //  stuff etc
    function Register: HResult; override; stdcall;

    function UnRegister: HResult; override; stdcall;

    // constructor and destructor
    constructor Create(AStream: TBCTStream; Amt: PAMMediaType;
      out hr: HResult);
  end;

  TStreamFilter = class(TComponent)  // IFileSourceFilter
  private
    { Private declarations }
    FFilterGraph   : TMyFilterGraph;
    FActive        : boolean;
    FStreamReader  : TBCTStreamReader;
    FGraphBuilder  : IGraphBuilder;
    FPin           : IPin;
    FRaiseInternalExceptions:boolean;
    FStream        : TStream;
    FOffset        : integer;
    procedure SetFilterGraph(AFilterGraph: TMyFilterGraph);
    procedure setActive(value:boolean);
    function GetFilterGraph:IGraphBuilder;
    procedure setOffset(value:integer);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure play;
    procedure setStream(reader:TStream; mediaType:TAMMediaType);
    function getPin:TBCBasePin;
    function getMediaType:TAMMediaType;
  published
    { Published declarations }
    { The @link(TMyFilterGraph) component }
    property FilterGraph: TMyFilterGraph read FFilterGraph write SetFilterGraph default nil;
    property Active:boolean read FActive write setActive default false;
    property RaiseInternalExceptions:boolean read FRaiseInternalExceptions write FRaiseInternalExceptions;
    property Offset:integer read FOffset write setOffset default 0;
  end;

  TFileStreamWrapper=class(TComponent)
  private
    FFileStream:TFileStream;
    FFileName:string;
    FActive:boolean;
    FStreamFilter:TStreamFilter;
    FMediaType:TAMMediaType;
    FDesiredExtension:string;
    procedure setFileName(value:string);
    procedure setActive(value:boolean);
    procedure setStreamFilter(value:TStreamFilter);
  public
    destructor Destroy; override;
  published
    property FileName:string read FFileName write setFileName;
    property DesiredExtension:string read FDesiredExtension write FDesiredExtension;
    property Active:boolean read FActive write setActive default false;
    property StreamFilter:TStreamFilter read FStreamFilter write setStreamFilter default nil; 
  end;

procedure Register;

implementation

uses Windows;

procedure Register;
begin
  RegisterComponents('Ciuly''s DSPack Filters', [TStreamFilter]);
  RegisterComponents('Ciuly''s DSPack Filters', [TMyFilterGraph]);
  RegisterComponents('Ciuly''s DSPack Filters', [TFileStreamWrapper]);
end;

function getMediaType(FileName:string; desiredExtension:string):TAMMediaType; 
Var  dsMediaType    : TAMMediaType;
     dsMediaStreams : Integer;
//     dsMediaLength  : Double;
     dsMediaDet     : IMediaDet;
     I              : Integer;
begin
  if desiredExtension='' then desiredExtension := UpperCase(ExtractFileExt(FileName))
                         else desiredExtension := UpperCase(desiredExtension);{}
  ZeroMemory(@result,sizeof(TAMMediaType));
  if desiredExtension='.OGG' then
  begin
    result.subtype:=MEDIASUBTYPE_Ogg;
    result.majortype:=WMMEDIATYPE_Audio;
    result.formattype:=FORMAT_VorbisFormat;
    exit;
  end;
 If FileExists(FileName)then
   If CoCreateInstance(CLSID_MediaDet, nil, CLSCTX_INPROC, IID_IMediaDet,dsMediaDet) = S_OK then
     If dsMediaDet.put_FileName(FileName) = S_OK then
       If dsMediaDet.get_OutputStreams(dsMediaStreams) = S_OK then
       Begin
//         dsMediaDet.
         If dsMediaStreams > 0 then
            For I := 0 to dsMediaStreams -1 do
               If dsMediaDet.put_CurrentStream(I) = S_OK then
                  If dsMediaDet.get_StreamMediaType(dsMediaType) = S_OK then
                    if dsMediaStreams=1 then
                    begin
                      if desiredExtension='.OGG' then
                      begin
                        result:=dsMediaType;
                        result.subtype:=MEDIASUBTYPE_Ogg;
                      end;
                    end;
       end;
end;

procedure TMyFilterGraph.InsertFilter(AFilter: IFilter);
begin
  inherited;
end;

procedure TMyFilterGraph.RemoveFilter(AFilter: IFilter);
begin
  inherited;
end;

//----------------------------------

constructor TBCTStream.Create(AStream:TStream; AOffset:integer);
begin
  if not assigned(AStream) then
  begin
    raise Exception.Create('TBCTStream.Create(AStream:TStream) : AStream cannot be unassigned (NIL)!');
    exit;
  end;
  FStream:=AStream;
  FOffset:=AOffset;
  FStream.Seek(FOffset,soFromBeginning);

  FLock := TBCCritSec.Create;
  Inherited Create;
end;

destructor TBCTStream.Destroy;
begin
  if Assigned(FLock) then
    FreeAndNil(FLock);
  Inherited Destroy;
end;

function TBCTStream.SetPointer(APos: LONGLONG): HResult;
begin
  if (APos < 0) or (APos > FStream.Size) then
    Result := S_FALSE
  else
    begin
      FStream.Seek(FOffset+APos,soFromBeginning);
      Result := S_OK;
    end;
end;

function TBCTStream.Read(ABuffer: PByte; ABytesToRead: DWord;
  AAlign: Boolean; out ABytesRead: DWord): HResult;
type
  TBuf=array of byte;
  PBuf=^TBuf;
var
  _ReadLength: DWord;
begin
  FLock.Lock;
  try
    if (FStream.Position + ABytesToRead > FStream.Size) then
      _ReadLength := FStream.Size - FStream.Position 
    else
      _ReadLength := ABytesToRead;

    ABytesRead := FStream.Read(PBuf(ABuffer)^,_ReadLength);
    Result := S_OK;
  finally
    FLock.UnLock;
  end;
end;

function TBCTStream.Size(out ASizeAvailable: LONGLONG): LONGLONG;
begin
  ASizeAvailable := FStream.Size-FOffset;
  Result := FStream.Size-FOffset;
end;

function TBCTStream.Alignment: DWord;
begin
  Result := 1;
end;

procedure TBCTStream.Lock;
begin
  FLock.Lock;
end;

procedure TBCTStream.Unlock;
begin
  FLock.UnLock;
end;

//----------------------------------

constructor TBCTStreamReader.Create(AStream: TBCTStream;
  Amt: PAMMediaType; out hr: HResult);
begin
  Inherited Create('Stream Reader', nil, AStream, hr);

  CopyMemory(@Fmt, Amt, SizeOf(TAMMediaType));
end;

function TBCTStreamReader.Register: HResult;
begin
  Result := S_OK;
end;

function TBCTStreamReader.UnRegister: HResult;
begin
  Result := S_OK;
end;

//----------------------------------

  function TStreamFilter.GetFilterGraph:IGraphBuilder;
  var f:IGraphBuilder;
  begin
    result:=nil;
    if not FFilterGraph.Active then exit;
    if (failed(TFilterGraph(FFilterGraph).QueryInterface(TGUID(IID_IGraphBuilder),f))) and (FRaiseInternalExceptions) then
      raise Exception.Create('TStreamFilter.GetFilterGraph:IGraphBuilder : Failed to retrieve GraphBuilder.');
    result:=f;
  end;
  
  procedure TStreamFilter.setOffset(value:integer);
  begin
    if value=FOffset then exit;
    if FActive then
    begin
      if RaiseInternalExceptions then
        raise Exception.Create('Cannot set stream offset while filter is active!')
                                 else
        exit;
    end;
    if FStreamReader<>nil then
    begin
      if RaiseInternalExceptions then
        raise Exception.Create('Cannot set stream offset after the stream was set!')
                                 else
        exit;
    end;
    FOffset:=value;
  end;
  
  procedure TStreamFilter.SetFilterGraph(AFilterGraph: TMyFilterGraph);
  begin
    if AFilterGraph = FFilterGraph then exit;
    if FFilterGraph <> nil then // the filter was deleted
    begin// if stream filter is set, then delete it from the filtergraph
      if FStreamReader<>nil then 
        if (failed(GetFilterGraph.RemoveFilter(FStreamReader))) and (FRaiseInternalExceptions) then
          raise Exception.Create('TStreamFilter.SetFilterGraph(AFilterGraph: TMyFilterGraph) : Failed to remove StreamReader.');
      FActive:=false;
    end;
    FFilterGraph := AFilterGraph;
    if FFilterGraph <> nil then // the filter is being added
      FGraphBuilder:=GetFilterGraph;
  end;

  procedure TStreamFilter.setActive(value:boolean);
  begin
    if value=FActive then exit;
    if value then
    begin
      if FStreamReader=nil then
        raise Exception.Create('TStreamFilter.setActive(value:boolean) : Input stream not set!');
      if FFilterGraph=nil then
        raise Exception.Create('TStreamFilter.setActive(value:boolean) : FilterGraph not set!');

      CoInitializeEx(nil, COINIT_APARTMENTTHREADED);
      //  Make sure we don't accidentally go away!
      FStreamReader._AddRef;

      if FGraphBuilder=nil then FGraphBuilder:=GetFilterGraph;
      if (failed(FGraphBuilder.AddFilter(FStreamReader,'CiulyCustomStreamReader'))) and (FRaiseInternalExceptions) then
        raise Exception.Create('TStreamFilter.setActive(value:boolean) : Failed to add stream reader filter.');
      FPin:=FStreamReader.getPin(0);
      if (failed(FGraphBuilder.Render(FPin))) and (FRaiseInternalExceptions) then
        raise Exception.Create('TStreamFilter.setActive(value:boolean) : Failed to render pin.');
    end        else
    begin
      if (FFilterGraph <> nil) and (FStreamReader<>nil) then
        FGraphBuilder.RemoveFilter(FStreamReader);
      CoUninitialize;
    end;
    FActive:=value;
  end;

  constructor TStreamFilter.Create(AOwner: TComponent);
  begin
    inherited Create(AOwner);
    FFilterGraph:=nil;
    FStreamReader:=nil;
  end;

  destructor TStreamFilter.Destroy;
  begin
    if FActive then setActive(false);
    if FStreamReader<>nil then
      FStreamReader._Release;
    FilterGraph := nil;
    inherited destroy;
  end;

  procedure TStreamFilter.play;
  var MC:IMediaControl;
  begin
    if not FActive then
    begin
      if FRaiseInternalExceptions then
        raise Exception.Create('TStreamFilter.play : Filter is not active.');
      exit;
    end;
    MC:=nil;
    if (failed(FFilterGraph.QueryInterface(IID_IMediaControl,MC))) and (FRaiseInternalExceptions) then
      raise Exception.Create('TStreamFilter.play : Failed to query FilterGraph for media control.');
    if MC<>nil then
      if (failed(MC.Run)) and (FRaiseInternalExceptions) then
        raise Exception.Create('TStreamFilter.play : Failed to run media control.');
  end;

  procedure TStreamFilter.setStream(reader:TStream; mediaType:TAMMediaType);
  var hr: HResult;
      mt: TAMMediaType;
  begin
    if reader=FStream then exit;
    FStream:=reader;

    if reader<>nil then
    begin
      mt:=mediaType;
      mt.majortype := MEDIATYPE_Stream;
      FStreamReader := TBCTStreamReader.Create(TBCTStream.Create(reader,FOffset), @mt, hr);
    end            else
      FStreamReader:=nil;
  end;

  function TStreamFilter.getPin:TBCBasePin;
  begin
    result:=FStreamReader.GetPin(0);
  end;

  function TStreamFilter.getMediaType:TAMMediaType;
  begin
    result:=FStreamReader.Fmt;
  end;

//----------------------------

procedure TFileStreamWrapper.setFileName(value:string);
begin
  if value=FFileName then exit;
  if FActive then
    raise Exception.Create('TFileStreamWrapper.setFileName(value:string) : Cannot set filename while active.');
  if (value<>'') and (not fileexists(value)) then
    raise Exception.Create('TFileStreamWrapper.setFileName(value:string) : the value specified does not denote an existing filename.');
  FFileName:=value;
  if value<>'' then
  begin
    FFileStream:=TFileStream.Create(FFileName,fmOpenRead or fmShareDenyWrite);
    FMediaType:=getMediaType(FFileName, FDesiredExtension);
  end          else
    if assigned(FFileStream) then FFileStream.Free;
end;

procedure TFileStreamWrapper.setActive(value:boolean);
var mt:TAMMediaType;
begin
  if value=FActive then exit;
  if value then
  begin
    if FFileName='' then
      raise Exception.Create('TFileStreamWrapper.setActive(value:boolean) : Must set filename first.');
    if not assigned(FStreamFilter) then
      raise Exception.Create('TFileStreamWrapper.setActive(value:boolean) : Must set StreamFilter first.');
    mt:=getMediaType(FFileName, FDesiredExtension);
    FStreamFilter.setStream(FFileStream,mt);
  end      else
  begin
    if assigned(FFileStream) then FFileStream.Free;
    FStreamFilter.setStream(nil,FMediaType);
  end;
  FActive:=value;
end;

procedure TFileStreamWrapper.setStreamFilter(value:TStreamFilter);
begin
  if value=FStreamFilter then exit;
  if FActive then
    raise Exception.Create('TFileStreamWrapper.setStreamFilter(value:TStreamFilter) : Cannot set StreamFilter while active.');
  if value=nil then
    FStreamFilter.setStream(nil,FMediaType);
  FStreamFilter:=value;
end;

destructor TFileStreamWrapper.Destroy;
begin
  try
    if assigned(FFileStream) then FFileStream.Free;
    if FStreamFilter<>nil then
      FStreamFilter.setStream(nil,FMediaType);
  except
  end;

  inherited;
end;

end.
